﻿#pragma TextEncoding = "UTF-8"
#pragma rtGlobals=3
#pragma version=1.20
#pragma igorVersion=7
#pragma IndependentModule=UserProcLoader

constant kFAST=1
// fast option is intolerant of semicolons in filenames

// https://www.wavemetrics.com/user/tony

// Provides an easy way to load and unload procedure files from the User 
// Procedures or Wavemetrics Procedures folders.

// Requires Igor Pro 8 for full functionality.
// In Igor 7 it's possible to select procedures that cannot be unloaded

// Trying to load a procedure that won't compile could lead to confusion.
// Version 1.20 allows following of shortcuts/aliases, and is tolerant of 
// semicolons in filenames. Repopulating the procedure lists, however, is
// significantly slower.

menu "Macros"
	"Load or Unload...",/Q, makeProcLoaderPanel()
end

function makeProcLoaderPanel()
	dowindow /k LoaderPanel
	NewDataFolder /O root:Packages
	NewDataFolder /O root:Packages:procLoader
	DFREF dfr=root:Packages:procLoader
	
	// prepare waves for two listboxes			
	make /o /n=0 dfr:w_loadedSelection /WAVE=w_loadedSelection, dfr:w_unloadedSelection /WAVE=w_unloadedSelection
	make /o /n=0/T dfr:w_loadedProcs, 	dfr:w_unloadedProcs, dfr:w_loadedTitle, dfr:w_unloadedTitle
	wave /SDFR=dfr/T w_loadedProcs, w_unloadedProcs, w_loadedTitle, w_unloadedTitle
	w_loadedTitle={"loaded"}; w_unloadedTitle={"unloaded"}
				
	// create panel	
	NewPanel /K=1/N=LoaderPanel/W=(20,20,520,380) as "Load or Unload Procedures"
	ModifyPanel /W=LoaderPanel fixedSize=1, noedit=1
	
	// popup to select procedures folder on disk
	popupMenu popupFolder, win=LoaderPanel, pos={20, 10}, value="User Procedures;Wavemetrics Procedures", proc=loaderPopMenuProc
	
	// button to open Procedure Browser window (useful if you want to open procedure after loading)
	button buttonBrowser, win=LoaderPanel, pos={460, 15}, size={15,14}, picture=thumb, labelBack=0, title="", proc=loaderButtonProc, help={"Open Procedure Browser"}

	// listbox for procedure files in folder that are not currently loaded
	ListBox listboxUnloaded,win=LoaderPanel, pos={20, 40},size={200, 300}, listwave=w_unloadedProcs, selwave=w_unloadedSelection, mode=9
	ListBox listboxUnloaded,win=LoaderPanel, fsize=12, proc=loaderListBoxProc, titleWave=w_unloadedTitle
	
	// buttons to be enabled when a valid selection is made from a listbox
	button buttonLoad win=LoaderPanel, pos={230,120},size={40,40},title=">", proc=loaderButtonProc, disable=2
	button buttonUnload win=LoaderPanel, pos={230,170},size={40,40},title="<", proc=loaderButtonProc, disable=2
	
	// listbox for procedure files in folder that ARE currently loaded
	ListBox listboxLoaded,win=LoaderPanel, mode=9, selwave=w_loadedSelection, pos={280, 40},size={200, 300}, editstyle=1
	ListBox listboxLoaded,win=LoaderPanel, listwave=w_loadedProcs,fsize=12, proc=loaderListBoxProc, titleWave=w_loadedTitle
	
	// populate listbox waves
	updateListboxWaves()
	return 1
end

function loaderListBoxProc(lba) : ListBoxControl
	STRUCT WMListboxAction &lba
	
	DFREF dfr=root:Packages:procLoader	
	
	if(datafolderRefStatus(dfr)==0)
		return 0
	endif
	
	if(lba.eventCode==-1) // clean up
		killDataFolder /Z dfr
		return 1
	endif
	
	string procWinStr=""
	
	if (lba.eventCode==4 && lba.eventMod&2^4) // right click in listbox cell
		if (stringmatch(lba.ctrlName, "listboxUnloaded"))
			return 0 // allow context menu for loaded files
		endif
		procWinStr=lba.listWave[lba.row]+".ipf"
		wave /T w_normalProcs=ListToTextWave(winlist("*",";","WIN:128,INDEPENDENTMODULE:0"),";")
		string popupStr
		if (isInTextWave(procWinStr,w_normalProcs)==0)
			// cell contains an independent module, user will have to find another way to open :)
			popupStr="(show "+ procWinStr
		else
			popupStr="show "+ procWinStr
		endif
		PopupContextualMenu popupStr
		if(V_flag)
			DisplayProcedure/W=$procWinStr
		endif
		return 0
	endif	
	
	if(lba.eventCode==3 )//&& lba.row>-1) // double click
		if (stringmatch(lba.ctrlName, "listboxUnloaded") && lba.row>-1)
			IncludeListboxSelection(1) // load procs in unloaded listbox on double click
		else // try to display the proc window on double click in loaded listbox
			procWinStr=lba.listWave[lba.row]+".ipf"
			DisplayProcedure/W=$procWinStr
			GetWindow /Z $procWinStr active
			if (V_Value==0) // didn't succeed in displaying proc
				doigorMenu "Windows", "Procedure Browser"
			endif
		endif
	endif	
	
	wave /SDFR=dfr w_loadedSelection, w_unloadedSelection
	
	// are we selecting procedures to load or to unload?
	variable load=stringmatch(nameofwave(lba.selWave), "w_unloadedSelection")
	
	if(lba.eventCode==4 || lba.eventCode==5)
		if(load && numpnts(w_loadedSelection))
			w_loadedSelection=w_loadedSelection & ~1 
			w_loadedSelection=w_loadedSelection & ~8 // clear selection in other listbox
		endif
		if(load==0 && numpnts(w_unloadedSelection))
			w_unloadedSelection=0
		endif
		// don't allow disabled cells to be selected
		// (can be done by shift-click or using arrow keys)
		lba.selWave = (lba.selWave&129)==129 ? lba.selWave & ~1 : lba.selWave
		lba.selWave = (lba.selWave&136)==136 ? lba.selWave & ~8 : lba.selWave
	endif
	
	// enable buttons based on selection	
	updateButtons()	
	return 0
end

function loaderPopMenuProc(pa)
	STRUCT WMPopupAction &pa

	if(pa.eventCode==2)
		updateListboxWaves()
	endif
	return 0
end

function loaderButtonProc(ba)
	STRUCT WMButtonAction &ba
	
	if(ba.eventCode!=2)
		return 0
	endif
	
	if(stringmatch(ba.ctrlName, "buttonBrowser"))
		doIgorMenu "Windows", "Procedure Browser"
	elseif(stringmatch(ba.ctrlName, "buttonLoad"))
		IncludeListboxSelection(1)
	elseif(stringmatch(ba.ctrlName, "buttonUnload"))
		IncludeListboxSelection(0)
	endif
	return 0	
end

function updateButtons()
	
	DFREF dfr=root:Packages:procLoader	
	wave /SDFR=dfr w_loadedSelection, w_unloadedSelection
		
	variable selections=0
	if (numpnts(w_unloadedSelection))
		selections=sum(w_unloadedSelection)
	endif
	button buttonLoad win=LoaderPanel, disable=2*(selections==0)
	// examine first bit of each point in w_loadedSelection
	selections=0
	if (numpnts(w_loadedSelection))
		duplicate /free w_loadedSelection w_testSel
		w_testSel=w_testSel&1	
		selections=sum(w_testSel)
	endif
	button buttonUnload win=LoaderPanel, disable=2*(selections==0)
end

function IncludeListboxSelection(insert)
	variable insert // 1 for InsertInclude, 0 for DeleteInclude
	
	DFREF dfr=root:Packages:procLoader
	if(insert)
		wave w_sel=dfr:w_unloadedSelection
		wave /T w_list=dfr:w_unloadedProcs
	else
		wave w_sel=dfr:w_loadedSelection
		wave /T w_list=dfr:w_loadedProcs
	endif
	string cmd=selectString(insert, "DELETEINCLUDE", "INSERTINCLUDE")
	string cmd2=selectString(insert, "unloaded", "loaded")	
	controlinfo /W=LoaderPanel popupFolder
	variable isWMproc=v_value-1
	string startquote=selectString(isWMproc, "\"", "<")
	string endquote=selectString(isWMproc, "\"", ">")
	string strExct="", procList=""
	variable i
	
	for(i=0;i<numpnts(w_sel);i+=1)
		if( (w_sel[i]&2^0) || (w_sel[i]&2^3) )
			sprintf strExct, "%s %s%s%s", cmd, startquote, w_list[i], endquote 			
			Execute/P/Q/Z strExct
			procList=addlistItem(w_list[i],procList)
		endif
	endfor
	sprintf cmd2, "#selectItems(\"%s\",\"%s\")", cmd2, procList
	Execute/P/Q/Z "COMPILEPROCEDURES "	
	Execute/P/Q/Z getindependentModuleName()+"#updateListboxWaves()" // queue an update for the listboxes	
	Execute/P/Q/Z getindependentModuleName()+cmd2 // select the items that have just been loaded/unloaded
	return 1	
end

// search recursively for ipf files
// returns list of files as textwave
// slow method avoids IndexedFile(path, -1, string) and IndexedDir(path, -1, string) 
// because semicolon is valid character filenames!
function /wave recursiveProcWave()
		
	variable maxFolders=1000
	variable isWin=stringmatch(StringByKey("OS", igorinfo(3))[0,2],"Win")
	variable folderIndex, i
	string strFile="", strFolder=""			
	make /free/T/n=0 w_folders, w_allFiles
	
	w_folders={SelectedFolder()} // search recursively starting in this folder

	// w_folders will grow as subfolders are added
	for(folderIndex=0;folderIndex<numpnts(w_folders);folderIndex+=1)
		// check files in folderIndexth folder from w_folders		
		newPath /O/Q/Z tempPathFF, w_folders[folderIndex]		
		
		if(kFast) // don't worry about semicolons in filenames, or shortcuts ending in .ipf!
			// add list of ipf files in current folder to w_allFiles
			wave /T w=ListToTextWave(IndexedFile(tempPathFF, -1, ".ipf"),";")
			TextWaveConcatenate(w_allFiles, w)
			// add list of folders in current folder to w_folders
			wave /T w=ListToTextWave(IndexedDir(tempPathFF, -1, 1),";")
			w+=":"
			TextWaveConcatenate(w_folders, w)
			
			if(isWin)
				// get a list of shortcuts 
				wave /T w=ListToTextWave(IndexedFile(tempPathFF, -1, ".lnk"),";")
				if(numpnts(w))
					w=ResolveAlias(w_folders[folderIndex]+w)
					duplicate /free w w_f 
					// keep shortcuts to files in w, shortcuts to folders in w_f
					w=PathEnd(w)
					TextWaveStringMatch(w, "*.ipf")
					TextWaveConcatenate(w_allFiles, w)
					TextWaveStringMatch(w_f, "*:")
					TextWaveConcatenate(w_folders, w_f)
				endif
			else // mac			
				// get a list of file aliases
				wave /T w=ListToTextWave(IndexedFile(tempPathFF, -1, "alis"),";")
				if(numpnts(w))
					w=ResolveAlias(w_folders[folderIndex]+w)
					w=PathEnd(w)
					TextWaveStringMatch(w, "*.ipf")
					TextWaveConcatenate(w_allFiles, w)
				endif
				// get a list of folder aliases
				wave /T w=ListToTextWave(IndexedFile(tempPathFF, -1, "fdrp"),";")
				if(numpnts(w))
					w=ResolveAlias(w_folders[folderIndex]+w)
					TextWaveZapBlanks(w)
					TextWaveConcatenate(w_folders, w)
				endif			
			endif
			continue // next folder
		endif
				
		// the slow way: loop through files in folder
		for(i=0;1;i+=1)
			strFile=IndexedFile(tempPathFF, i, "????")		
			if(strlen(strFile)==0)
				break
			endif
			GetFileFolderInfo /Q/Z w_folders[folderIndex]+strFile
			if(V_isAliasShortcut)
				strFile=ResolveAlias(w_folders[folderIndex]+strFile)
				if(stringmatch(strFile,"*:"))
					w_folders[numpnts(w_folders)]={strFile}
					continue // next file
				endif
				strFile=ParseFilePath(0, strFile, ":", 1, 0)
			endif			
			if(stringmatch(strFile,"*.ipf"))
				w_allFiles[numpnts(w_allFiles)]={strFile}
				continue
			endif
		endfor // next file
		
		// add subfolders in current folder
		for(i=0;1;i+=1)
			strFolder=IndexedDir(tempPathFF, i, 1)
			if(strlen(strFolder)==0)
				break
			endif
			w_folders[numpnts(w_folders)]={strFolder+":"}
		endfor // next subfolder
		if(folderIndex==maxFolders)
			print "Warning: RecursiveProcWave exceeding max iterations"
			break
		endif		
	endfor // next folder
	
	killpath /Z tempPathFF
	return w_allFiles
end


// because w = ParseFilePath(0, w, ":", 1, 0) doesn't compile ???
function /s PathEnd(s)
	string s
	
	if (numtype(strlen(s))==2)
		return ""
	endif	
	return ParseFilePath(0, s, ":", 1, 0)
end

// appends w2 to w
function TextWaveConcatenate(w, w2)
	wave /T w, w2
	
	variable oldPnts=numpnts(w), newPnts=numpnts(w2)	
	if(newPnts)
		redimension /n=(oldPnts+newPnts), w
		w[oldPnts,]=w2[p-oldPnts]
	endif
end

// keep only points of textwave matching strMatch
function TextWaveStringMatch(w, strMatch)
	wave /T w
	string strMatch
	
	if(numpnts(w))
		w=selectstring(stringmatch(w,strMatch),"",w)
		TextWaveZapBlanks(w)
	endif	
end

// Recursively resolve shortcuts to files/directories
// returns full path or an empty string if the file does not exist or the
// shortcut points to a non existing file/folder
function/S ResolveAlias(strPath)
	string strPath // full path to file, folder or alias

	GetFileFolderInfo/Q/Z removeEnding(strPath, ":")
	if(V_isAliasShortcut)
		return ResolveAlias(S_aliasPath)
	endif
	if(v_flag==0) 
		return strPath // path to a file or folder
	endif
	return ""
end

// returns path to folder selected in popup
// 1: User Procedures 2: Wavemetrics Procedures
function /S SelectedFolder()
	
	controlinfo /W=LoaderPanel  popupFolder
	if (V_flag<1)
		return ""
	endif
	if(v_value==1)
		return SpecialDirPath("Igor Pro User Files",0,0,0)+"User Procedures:"
	endif
	return SpecialDirPath("Igor Application",0,0,0)+"Wavemetrics Procedures:"
end

// strip trailing stuff in the output from winlist
function /S filenameFromString(s)
	string s
	
	if (numtype(strlen(s))==2)
		return ""
	endif
	
	SplitString/E=".*\.ipf" s // remove module names
	return removeending(s_value, ".ipf")
end

// extract quoted filename from include statement
function /S filenameFromIncludeStatement(s)
	string s

	if (numtype(strlen(s))==2)
		return ""
	endif
	
	controlinfo /W=LoaderPanel popupFolder
	variable isWMproc=v_value-1
	string startquote=selectString(isWMproc, "\"", "<")
	string endquote=selectString(isWMproc, "\"", ">")
	
	variable start, stop
	start=strsearch(s, startquote, 0)
	stop=strsearch(s, endquote, start+1)
	if (start==-1 || stop==-1)
		return ""
	endif
	return s[start+1,stop-1]
end

// truth that string s is a member of textwave w
function isInTextWave(s, w)
	string s
	wave /T w
	
	if(numtype(strlen(s))==2)
		return 0
	endif
	
	FindValue /TEXT=s/TXOP=4 /Z w		
	return (V_Value>-1)
end

// remove empty points from textwave w
function TextWaveZapBlanks(w)
	wave /T w
	
	if(numpnts(w)==0)
		return 0
	endif
	variable i=0
	do
		if(stringmatch(w[i],""))
			deletepoints i,1,w
		else	
			i+=1
		endif
	while(i<numpnts(w))
	return numpnts(w)
end

function updateListboxWaves()
	
	DFREF dfr=root:Packages:procLoader
	wave /SDFR=dfr w_loadedSelection, w_unloadedSelection
	variable numLoaded, numUnloaded, numIncluded
	
	// create a free text wave containing a list of procedure files in the selected source folder	
	wave /T w_unloadedProcs=RecursiveProcWave()
	numUnloaded=numpnts(w_unloadedProcs)	
	
	// create a free text wave containing a list of all procedure files in the experiment
	// use colon as list separator because semicolon is legal in filenames
	wave /T w_loadedProcs=ListToTextWave(winlist("*",":","WIN:128,INDEPENDENTMODULE:1"),":")	
	numLoaded=numpnts(w_loadedProcs)
	
	// clean up names
	if(numLoaded)		
		w_loadedProcs=FilenameFromString(w_loadedProcs) // some points may be blank
	endif
	if(numUnloaded)
		w_unloadedProcs=removeending(w_unloadedProcs, ".ipf")
	endif
	
	// remove procedures that aren't in selected source folder from loaded list
	if(numLoaded)		
		w_loadedProcs=selectstring(isInTextWave(w_loadedProcs[p], w_unloadedProcs),"",w_loadedProcs[p])
	endif
	
	// remove already loaded procedures from unloaded list
	if(numUnloaded)
		w_unloadedProcs=selectstring(isInTextWave(w_unloadedProcs[p], w_loadedProcs),w_unloadedProcs[p],"")
	endif
	
	// figure out which procedures can be unloaded.
	// using INCLUDE:2 with winlist gets all included files, 
	// not just those that we can DELETEINCLUDE
	
	// load procedure file lines into text wave
	wave /T w_includedProcs=ListToTextWave(ProcedureText("",0,"procedure"), "\r")
	
	// find lines that start with #include
	grep /E="^#include[\s]*[<\"]"/Z w_includedProcs as w_includedProcs
	numIncluded=TextWaveZapBlanks(w_includedProcs)
	
	// extract file names from between quotes
	if(numIncluded)
		w_includedProcs=filenameFromIncludeStatement(w_includedProcs)
	endif
  	
  	numUnloaded=TextWaveZapBlanks(w_unloadedProcs)
  	numLoaded=TextWaveZapBlanks(w_loadedProcs)
	
  	sort w_loadedProcs, w_loadedProcs
	sort w_unloadedProcs, w_unloadedProcs
	
	// copy free waves into the package folder
	duplicate /o/T w_unloadedProcs dfr:w_unloadedProcs /wave=w_unloadedProcs
	duplicate /o/T w_loadedProcs dfr:w_loadedProcs /wave=w_loadedProcs
	
	redimension /N=(numLoaded) w_loadedSelection
	redimension /N=(numUnloaded) w_unloadedSelection
	
	// disable cells containing procedures that cannot be unloaded using DELETEINCLUDE	
	if(numLoaded)
		w_loadedSelection=(1-isInTextWave(w_loadedProcs, w_includedProcs))*2^7
	endif
	if(numUnloaded)
		w_unloadedSelection=0
	endif	
	
	// nothing is selected, so disable load and unload buttons
	button /Z buttonLoad win=LoaderPanel, disable=2
	button /Z buttonUnload win=LoaderPanel, disable=2
	return 1
end

// called after updateListboxWaves() to select the listbox items in selList
function selectItems(strBox, selList)
	string strBox, selList // strBox: "loaded" or "unloaded"
	
	DFREF dfr=root:Packages:procLoader	
	if(stringmatch(strBox, "loaded"))
		wave selWave=dfr:w_loadedSelection
		wave /T listWave=dfr:w_loadedProcs
	else
		wave selWave=dfr:w_unloadedSelection
		wave /T listWave=dfr:w_unloadedProcs	
	endif
	
	variable i, numItems
	numItems=itemsinlist(selList)
	for (i=0;i<numItems;i+=1)
		findValue /TEXT=stringFromList(i, selList) /Z listWave
		if(V_value>-1)
			selWave[v_value]=1
		endif
	endfor
	ListBox $"listbox"+strBox,win=LoaderPanel, row=-1
	updateButtons()
end

// I used a screenshot to steal the procedure browser button from the control bar!
// make non-uniform grey background at edge of button transparent:
// •thumb[28,][][3] = (thumb[p][q][0]==thumb[p][q][1] && thumb[p][q][0]==thumb[p][q][2]) ? 0 : thumb[p][q][r]
// •thumb[][24,][3] = (thumb[p][q][0]==thumb[p][q][1] && thumb[p][q][0]==thumb[p][q][2]) ? 0 : thumb[p][q][r]
// create stacked images for button:
// •redimension /N=(90,-1,-1) thumb 
// •thumb[30,59][][]=thumb[mod(p,30)][q][r]
// then use Pictures dialog (misc - pictures) to create proc picture

// PNG: width= 90, height= 28
Picture thumb
	ASCII85Begin
	M,6r;%14!\!!!!.8Ou6I!!!"&!!!!=#R18/!9'oUErZ1?&TgHDFAm*iFE_/6AH5;7DfQssEc39jTBQ
	=U#OG#M5u_NK`L;'g'_`?_5JPn)_d:C()V]%S0h9%3pu4-+e;tk,")Zj,2$t(u:^^!26PP,C#]8!<9
	mWaQ8X-SSBWG/hC(;G*#S&af\J(Vo(h3i=JdEa20h9%Bm_&NRbPIltcZl$<dZ3HKn$CpGHi)smgJt<
	2g*[<'Xa;<t!BdsaPsOO"h=u=#@$P%=i-3?c*Ydi'%Ab)j+3>HX.,7PocZV:W6E^l,aImEkaM]:*'m
	bL`L6pfSJJc63O+3ANeuTl'kA,hOWe93S-G[q*-P6FrcH'TnrT4.>cLlR_KAXsdKjN",$]&0nO+B&R
	">RL]CN;`9._&d][6h'7(^>`1g1:YEMd]WVWriKZkEi[R2%1^O9gfP#?=&qC%a9eZ(_EhMq:"*r[P(
	KYl.*a,BQsH@RWK&3?uap)g+'eI@0A8Pps!/h+sK/DUP6@0D;B_m<dJ-;i7BXaD/V`L"ShK%.6s9po
	k[6<D&EoQbc$+'I\kk\>,>;ai(")=K[B&aRSE%6V,0L-n&sp<gjpr?RaQf%,in^6N4(uiYXT#H/=d<
	="ari;0Dbn$'\G6r;P6kA]QME!=TX0"2;/TMK\=g09eO;AaD^g4$Rg.Y$Or)TjQC*,J.Z>J1TU05@M
	f319f>^S[jh"m35c&+[WA_9$^@M3L;Cr>M0ud?+mCG\D)[i!6,-,m$C_6lSZh0D=ZLU`^-jE(L_/a+
	@f3_e&j>Z+=DRP`\rK#QD&4D5Whg)JX6sW6K#Te`(hHA'h=\a7nTbMA)&$++2'5n%3),AA>ckQTTDn
	[.T'-Fq5Ptp$CkW#c*j.arR/OEDrcr-lYIG0j,k7=iA&VI'^(DJ2.>=jc*pe%TXfX21WitUGK?%te4
	%&aEV)R%LR&uYtK$hc9'lJ(K_DRcVlV"d84.NhbMo_HN78f(sE/-nMMo!p_PPO:X(_.>Mk+uG"-8.R
	WPVW-g@/EC]CSaC>1;k6[M;iS52`f"sQc\`]A>7m@%\'1=+PaWqkMMU>3+#\%bW-',UhH[38oH,Q="
	_UMT\pp3]\I#$;cf3KSY5,^QcpuT(,I#S_&),;?$A@f7XJEo4d:R20_fV17S<CsP'_WLbc,36(1<TC
	k;?Gpi(D7*l>)m1(cR/r&,M(gf0e.1\fs>\Nr;6cL'X^#Nq23XINefNoU6ZRT;&.r>INp.<*Bdb3LR
	g=leH?k^JTssf?)=[jEsr;H]%PBL@fE8TbM)"'uYHIpj>ZC*B%+!p\T"nD"rBLe6_`D=nNZ3^)<7#%
	;#1$Kt@qJT.\7Kn$cWe0fM1',%mr013<Fogs5p1]k;5\:80CqI:\0FYp0=H7k5//)cJLPgpjK,ggo"
	jY1;'bQ8o#EDPObMCT^UF&Ja]lCeZcfkm_3GW+P^2\!rjjEP'+B6tgBGTPFn2<9UoLi"+*5h!Qkb>P
	s<*&L@,dRGb.&RD1V0_s+rJTeuh=B!thc/qRN!\f=HMIX_?uoS,-tn\t)"ZY"XcJX:ZkNH!U-N;/4k
	L:]G4-HpC!W2FJsCN.uBbmek'%s-+MBKcj&k@TCW\ZYfWHsN.5f/V]i`8B"%h.C[:"&KWihV1-(:#U
	#)`o5:L8oMIW*WCSN0fG<^$g2NTai!4^FLeiZ\V5qX/hh$nO1NeAqUNmr"Hno\W2m$h+$1CQDX2=[X
	$8jTD(Tb41&[\doMohnS)K*Gp)kP1Ir$1p2Z0:>:k`A30o?;/LU5sih&JJ'N.Uu?0%ZGS*ol%D10,u
	6bi)KSLqs""Va2GDEX'Qf[eEV$@YhsbOr;Goh%a1n^;RP*;-g?-3`jRUs&6h:l+('.88kD/L\/*K<@
	HiI?G(#$s*3QKE\^+jFk\g_mf"%tr'Z;(TgKJXRs1.'D>SMI->,)ZdWe3"W-"*ToAiq?lB+KniR<3f
	n0KAW?=$BP`As/"\EpZNTu>krVp8.gZCFe0PBm,7ou\bt\fp2G0E'`I^jJBuEkU*WfVrXG[/auCY9M
	heaJ?pV\9N9LKc:C[GiUhe)<lC?,H/A&LJt+O\5[m2Of<&r=S;h#k%mCQQo8r#8G!&^_JZoq9qa)rK
	2(h)J3/Ws_a[I(?cJ5'7laq)RH,VHmVitd)k-R"l8@kmXt_3<bI5[aRp:mDp1`+:ZF<+E4c,u3A[6;
	c+[]1_Z$Q),m+kUYi<+G;Fo)YGOV*OH:LhN+kBZZiI;M`<hY.o]G!QUZ7hP2ak3+0_n9AE11OCH].P
	%J^?U\I''4EL0M2>Q9k2oV=C"n]T<p<VFY'IY"=JZ]\?4R^tOt$>#)[m`ARae]<8r@fMNeg#VN:#3:
	+.li7!9QK=<%k8*_@me^MDp9b#PR*o&l6/!M6iJFE&b`X,-b>ga,15K^H#Dl='`Ri]r0YI/RPb\ieO
	Bd-9[9M4kgk=MrX!dSlOV(fk.LoZ/%(uAE,'!8n:Ts$&Qc'SN]VM9E`=]MgE@M9!@+MP)f7SP>=igc
	r<olZR@9f)mh7V0mBWc7)bQ[Y3]ASqdKZ(=6o5_QS2\h0farNU)f1d*h4>TeCWS#.b\A:3I5Xr$j3E
	:f7@8?k%T]s^!E))hE_N)5lR9V<)jpY)]LYJ6&'s>q`j;pX7,UPN#_lu!!#SZ:.26O@"J
	ASCII85End
End